# Zielfunktion an Stelle x
func1 <- function(x) {
    innerSum <- 0
    for(i in 1:length(x)) {
        innerSum <- innerSum + sum(x[1:i])^2
    }
    return(innerSum)
}


# Gradient der Zielfunktion an einer Stelle x
grad1 <- function(x, func=NULL, stepSize=NULL) {
    gradient <- c()
    gradient[length(x)] <- 2 * sum(x)
    for(i in (length(x)-1):1) {
        gradient[i] <- gradient[i+1] + 2 * sum(x[1:i])
    }
    return(matrix(gradient, ncol=1))
}


# optimiere die Zielfunktion für einen beliebigen Startpunkt
# Argumente:
# func - Zielfunktion
# grad - Gradientenfunktion
# startPoint - Startpunkt
# optStep - Funktion zum durchfuehren eines Schrittes
# initialStepSize - initiale Schrittweite
# maxIterations - maximale Anzahl Iterationen
# callback - Funktion zum Protokollieren des Verlaufs
optimizer <- function(func, grad, startPoint, optStep, initialStepSize, shrinkFactor=0.5, maxIterations=Inf, callback=NULL) {
    x <- startPoint
    currentF <- func(x)
    oldX <- x
    oldF <- currentF
    stepDiff <- Inf
    stepSize <- initialStepSize
    
    iteration <- 0
    while(iteration <= maxIterations) {
        oldX <- x
        oldF <- currentF
        
        # Durchführung des Verbesserungsschritts
        x <- optStep(x, func, grad, stepSize)
        currentF <- func(x)
        
        # evtl. protokollieren
        if (is.function(callback)) {
            callback(iteration, x, currentF, stepSize)
        }
        
        stepDiff <- oldF - currentF
        
        # Wenn der neue Funktionswert schlechter ist als der alte,
        # Schrittweite reduzieren, Loesung verwerfen
        if (stepDiff < 0) {
            stepSize <- stepSize * shrinkFactor
            x <- oldX
            currentF <- oldF
        }
        
        iteration <- iteration + 1
    }
    return(x)
}


# Diese Funktion führt einen Schritt im Gradientenabstieg durch
optimizationStep <- function(x, func, grad, stepSize) {
    g <- grad(x, func, stepSize)
    x - (stepSize %*% t(g / norm(g, type="f")))
}


# Hilfsfunktion zum Protokollieren
logFunction <- function(iteration, x, currentF, stepSize) {
    cat(iteration, x, currentF, stepSize, "\n")
}


## Ausfuehrung ##################################################
startPoint <- 10.0 + c(-0.11515555, 1.06158289, 0.06922904, 2.47318629, -0.35627440)
initialStepSize <- 10.0
shrinkFactor <- 0.5
startTime <- proc.time()[1]
result <- optimizer(func1, grad1, startPoint, optimizationStep, initialStepSize, shrinkFactor, 50, logFunction)
endTime <- proc.time()[1]
consumedTime <- as.numeric(endTime - startTime)
print(consumedTime)
print(func1(result))

